/* -------------------------------------------------------
                           hx1838.c

     Softwaremodul zum HX1838 IR-Receiver

     Belegt zur Auswertung des Receivers Timer1, IR-
     Receiver loest am angeschlossenen Pin einen
     externen Interrupt aus

     MCU   :  STM8S103F3
     Takt  :  interner Takt 16 MHz

     26.07.2017  R. Seelig
   ------------------------------------------------------ */

#include "hx1838.h"

volatile uint16_t  ir_code;                                                 // Code des letzten eingegangenen 16-Bit Wertes
volatile uint8_t   ir_newflag;                                              // zeigt an, ob ein neuer Wert eingegangen ist

/* --------------------------------------------------
                      tim1_init

     setzt Timer1 als Zaehler mit 100uS Takt
   --------------------------------------------------_ */
volatile void tim1_init(void)
{
  #define taktteilerpol     1599

  // Prescaler setzen
  // 0x063f  = 1599
  // entspricht bei 16 MHz einem Takteingang des Zaehlers mit 10000 Hz = 100uS


  TIM1_PSCRH= taktteilerpol >> 8;
  TIM1_PSCRL= taktteilerpol & 0x00ff;

  TIM1_CR1 |= TIM1_CR1_CEN;     // Timer1 clock enable
}

/* --------------------------------------------------
                      waittil_hi

     wartet, bis die Signalleitung zu logisch 1
     wird.

     Uebergabe:
        timeout  : Zeit x 100us nach der nach nicht
                   eingehen eines Hi-Pegels ein
                   Timeout ausgeloest wird
   -------------------------------------------------- */
uint8_t waittil_hi(uint16_t timeout)
{
  tim1_clr();
  while ( (!(is_irin() )) && (timeout > tim1_getvalue()) );
  if (timeout > tim1_getvalue()) return 0; else return 1;
}

/* --------------------------------------------------
                      waittil_lo

     wartet, bis die Signalleitung zu logisch 0
     wird.

     Uebergabe:
        timeout  : Zeit x 100us nach der nach nicht
                   eingehen eines Lo-Pegels ein
                   Timeout ausgeloest wird
   -------------------------------------------------- */
uint8_t waittil_lo(uint16_t timeout)
{
  tim1_clr();
  while ( (is_irin() ) && (timeout > tim1_getvalue()) );
  if (timeout > tim1_getvalue()) return 0; else return 1;
}

/* --------------------------------------------------
                      ir_getbit

     liefert je nach Pulselaenge des IR-Receivers 0
     (kurzer Impuls) oder 1 (langer Impuls). Bei
     Timeoutueberschreitung wird 2 zurueck geliefert

     Uebergabe:
        timeout  : Zeit x 100us nach der nach nicht
                   eingehen eines Lo-Pegels ein
                   Timeout ausgeloest wird
     Rueckgabe   : 0,1  Datenbit
                     2  Timeout aufgetreten
   -------------------------------------------------- */
uint8_t ir_getbit(uint16_t timeout)
{
  volatile uint16_t t;

  tim1_clr();
  while ( (is_irin() ) && (timeout > tim1_getvalue()) );
  t= tim1_getvalue();
  if (timeout <= t) return 2;
  if (t< 9) return 0; else return 1;
}

/* --------------------------------------------------
                  PortC_extint

     ISR fuer externe Interrupts auf PortC

     Dauer Lo-Pegel vor Startbit:   9 ms
     Startbit (Hi)              : 4.5 ms
     Datenbit Lo                : 0.6 ms
     Datenbit Hi                : 1.7 ms

     In der Interruptroutine werden die globalen
     Variablen ir_code und ir_newflag geschrieben,
     die in einem Hauptprogramm gepollt werden
     koennen.
   -------------------------------------------------- */
void PortC_extint(void) __interrupt(_GPIOC_VECT)
{
  volatile uint8_t cx, b, hw;
  volatile uint16_t result;

  if (!(is_irin()) )                                   // ist der Datenpin des IR-Receivers zu 0 geworden
  {
    ir_int_disable();                                  // Interruptfaehigkeit des Pins aus

    if (waittil_hi(120) ) goto timeout_err;            // auf Startbit des Frames warten (nach 12 ms Timeout)
    if (waittil_lo(60)  ) goto timeout_err;            // auf Ende Startbit des Frames warten (nach 6 ms Timeout)
    if (waittil_hi(30)  ) goto timeout_err;            // auf erstes Datenbit warten (nach 3 ms Timeout)

    // die ersten 16 Bit des Pakets einlesen ohne diese
    // auszuwerten
    for (cx= 0; cx < 16; cx++)
    {
      b= ir_getbit(30);
      if (b == 127) goto timeout_err;
      if (waittil_hi(30)  ) goto timeout_err;            // auf naechstes Datenbit warten
    }

    // erste 8 Bit seriell lesen
    hw= 0;
    for (cx= 0; cx < 8; cx++)
    {
      b= ir_getbit(30);
      if (b == 127) goto timeout_err;
      hw |= (b << cx);
      if (waittil_hi(30)  ) goto timeout_err;            // auf naechstes Datenbit warten
    }
    result= (uint16_t) hw << 8;

    // zweite 8 Bit seriell lesen
    hw= 0;
    for (cx= 0; cx < 8; cx++)
    {
      b= ir_getbit(30);
      if (b == 127) goto timeout_err;
      hw |= (b << cx);
      if (waittil_hi(30)  ) goto timeout_err;            // auf naechstes Datenbit warten
    }
    result |= hw;

    if (result < 0x8000)
    {
      ir_code= result;
      ir_newflag= 1;
    }

    timeout_err:

      delay(20);
      ir_int_enable();
      delay(20);
  }
}

/* --------------------------------------------------
              void portcint_init(void)

     initialisiert den Datenanschluss des IR-Receivers
     als Interrupt
   -------------------------------------------------- */
volatile void portcint_init(void)
{
  EXTI_CR1 &= ~(3 << EXTI_CR1_PCIS);     // external Interruptconditions fuer PortC loeschen...
  EXTI_CR1 |= (2 << EXTI_CR1_PCIS);      // ... um dann Conditions fuer Interupt bei neg. Flanke zu setzen

  int_enable();                          // Interrupts einschalten

  ir_input_init();
}
