/* ----------------------------------------------------------
                          uart_printf.c

     Softwaremodul fuer serielle Schnittstelle und redu-
     ziertem PRINTF fuer AVR-MCU. Protokoll 8N1


     MCU : STM8

     16.2.2019     R. Seelig

   ---------------------------------------------------------- */


#include "uart_printf.h"

#define echo_enable    0

char printfkomma = 1;


  #if (use_usart == 1)

    #define USARTx_SR           USART1_SR
    #define USARTx_DR           USART1_DR
    #define USARTx_BRR1         USART1_BRR1
    #define USARTx_BRR2         USART1_BRR2
    #define USARTx_CR1          USART1_CR1
    #define USARTx_CR2          USART1_CR2
    #define USARTx_CR3          USART1_CR3
    #define USARTx_CR4          USART1_CR4
    #define USARTx_CR5          USART1_CR5
    #define USARTx_GTR          USART1_GTR
    #define USARTx_PSCR         USART1_PSCR

  #endif


  #if (use_usart == 2)

    #define USARTx_SR           USART2_SR
    #define USARTx_DR           USART2_DR
    #define USARTx_BRR1         USART2_BRR1
    #define USARTx_BRR2         USART2_BRR2
    #define USARTx_CR1          USART2_CR1
    #define USARTx_CR2          USART2_CR2
    #define USARTx_CR3          USART2_CR3
    #define USARTx_CR4          USART2_CR4
    #define USARTx_CR5          USART2_CR5
    #define USARTx_GTR          USART2_GTR
    #define USARTx_PSCR         USART2_PSCR

  #endif


void uart_putchar(char ch)
{
  while(!(USARTx_SR & USART_SR_TXE));   // warten bis letztes Zeichen gesendet ist
  USARTx_DR = ch;                       // Zeichen senden
}


void uart_putstring(char *p)
{
  do
  {
    uart_putchar( *p );
  } while( *p++);
}



/* ------------------------------------------------------------
                         UART_GETCHAR
     wartet so lange, bis auf der seriellen Schnittstelle
     ein Zeichen eintrifft und liest dieses ein
   ------------------------------------------------------------ */
char uart_getchar(void)
{
  while(!(USARTx_SR & USART_SR_RXNE));    // solange warten bis ein Zeichen eingetroffen ist
  return USARTx_DR;
}

/* ------------------------------------------------------------
                         UART_ISCHAR
      testet, ob auf der seriellen Schnittstelle ein Zeichen
      zum Lesen ansteht (liest dieses aber nicht und wartet
      auch nicht darauf)
   ------------------------------------------------------------ */
char uart_ischar(void)
{
  return (USARTx_SR & USART_SR_RXNE);
}


/* ------------------------------------------------------------
                            UART_INIT
      Initialisieren der seriellen Schnittstelle. Mit internem
      Taktgeber funktioniert die Schnittstelle nur bis
      57600 Baud
   ------------------------------------------------------------ */

void uart_init(void)
{

  // Taktinitialisierung erfolgt hier, damit im MAIN ausser einem
  // uart_init() nichts gemacht werden muss, damit das System
  // funktioniert

  sysclock_init(0);                     // interner Taktgeber

  PD_DDR |= (1 << PD4);                                 // TxD Leitung als Output
  PD_CR1 |= (1 << PD4);

  USARTx_CR2 |= USART_CR2_TEN;                          // TxD enabled
  USARTx_CR2 |= USART_CR2_REN;                          // RxD enable
  USARTx_CR3 &= ~(USART_CR3_STOP1);                     // Stopbit

  // Baudratendivisor:
  //   Divisior = F_CPU / Baudrate
  //   Nibbles des Divisors N3 N2 N1 N0
  //   Bsp.:   16 MHz / 9600 bd = 1667d = 0683h
  //      Hier waere dann:
  //        N3 = 0   / N2 = 6  / N1 = 8  / N0 = 3

  //      BRR1 bekaeme N2 und N1 => 0x68
  //      BRR2 bekaeme N3 und N0 => 0x03

  //  USARTx_BRR1 = 0x68; // 9600 Baud
  //  USARTx_BRR2 = 0x03;

  if (BAUDRATE == 57600)
  {
    USARTx_BRR1 = 0x11;
    USARTx_BRR2 = 0x06;
  }
  if (BAUDRATE == 38400)
  {
    USARTx_BRR1 = 0x1a;
    USARTx_BRR2 = 0x0a;
  }
  if (BAUDRATE == 28800)
  {
    USARTx_BRR1 = 0x22;
    USARTx_BRR2 = 0x0c;
  }
  if (BAUDRATE == 19200)
  {
    USARTx_BRR1 = 0x34;
    USARTx_BRR2 = 0x02;
  }
  if (BAUDRATE == 9600)
  {
    USARTx_BRR1 = 0x68;
    USARTx_BRR2 = 0x03;
  }
  if (BAUDRATE == 4800)
  {
    USARTx_BRR1 = 0xd0;
    USARTx_BRR2 = 0x05;
  }
}


#if (readint_enable == 1)

  /* ------------------------------------------------------------
                              readint

       liest einen 16-Bit signed Integer auf der seriellen
       Schnittstelle ein (allerdings reicht der Eingabebereich
       nur von -32767 .. +32767).

       Korrektur ist mit der Loeschtaste nach links moeglich.
     ------------------------------------------------------------ */
  int16_t readint(void)
  {
    uint8_t   signflag= 0;
    uint16_t  sum = 0;
    uint8_t   zif;
    uint8_t   ch;

    do
    {
      ch= uart_getchar();
      if (ch== 0x0a) ch= 0x0d;

      // Ziffern auswerten
      if ((ch>= '0') && (ch<= '9'))
      {
        zif= ch-'0';

        if ((sum== 0) && (zif))        // erste Ziffer
        {
          sum= zif;
          uart_putchar(ch);
        }
        else
        if (sum < 3277)
        {
          if  (!( ((sum * 10) > 32750) && (zif > 7) ))
          {
            sum= (sum*10) + zif;
            uart_putchar(ch);
          }
        }
      }

      // letzte Eingabe loeschen
      if ((ch== 127) || (ch== 8))    // letzte Eingabe loeschen
      {
        if (sum)
        {
          sum /= 10;
          uart_putchar(8);
          uart_putchar(' ');
          uart_putchar(8);
        }
        else
        if (signflag)
        {
          uart_putchar(8);
          uart_putchar(' ');
          uart_putchar(8);
          signflag= 0;
        }
      }

      // Eingabe Minuszeichen
      if ((ch== '-') && (sum == 0))
      {
        signflag= 1;
          uart_putchar('-');
      }

    } while (ch != 0x0d);              // wiederholen bis Returnzeichen eintrifft
    if (signflag) return sum *(-1); else return sum;
  }

#endif

/* ------------------------------------------------------------
                            PUTINT
     gibt einen Integer dezimal aus. Ist Uebergabe
     "komma" != 0 wird ein "Kommapunkt" mit ausgegeben.

     Bsp.: 12345 wird als 123.45 ausgegeben.
     (ermoeglicht Pseudofloatausgaben im Bereich)
   ------------------------------------------------------------ */
void putint(int16_t i, char komma)
{
  typedef enum boolean { FALSE, TRUE }bool_t;

  static uint16_t zz[]      = { 10000, 1000, 100, 10 };
  bool_t          not_first = FALSE;

  uint8_t    zi;
  int16_t    z, b;

  komma= 5-komma;

  if (!i)
  {
    uart_putchar('0');
  }
  else
  {
    if(i < 0)
    {
      uart_putchar('-');
      i = -i;
    }

    for(zi = 0; zi < 4; zi++)
    {
      z = 0;
      b = 0;

      if  ((zi== komma) && komma)
      {
        if (!not_first) uart_putchar('0');
        uart_putchar('.');
        not_first= TRUE;
      }

      while(z + zz[zi] <= i)
      {
        b++;
        z += zz[zi];
      }
      if(b || not_first)
      {
        uart_putchar('0' + b);
        not_first = TRUE;
      }
      i -= z;
    }
    if (komma== 4) uart_putchar('.');
    uart_putchar('0' + i);
  }
}

/* ------------------------------------------------------------
                       HEXNIBBLEOUT
     gibt die unteren 4 Bits eines chars als Hexaziffer aus.
     Eine Pruefung ob die oberen vier Bits geloescht sind
     erfolgt NICHT !
  -------------------------------------------------------------  */
void hexnibbleout(uint8_t b)
{
  if (b< 10) b+= '0'; else b+= 55;
  uart_putchar(b);
}

/* ------------------------------------------------------------
                            PUTHEX
     gibt einen Integer hexadezimal aus. Ist die auszugebende
     Zahl >= 0xff erfolgt die Ausgabe 2-stellig, ist sie
     groesser erfolgt die Ausgabe 4-stellig.
   ------------------------------------------------------------ */
void puthex(uint16_t h)
{
  uint8_t b;

  if (h> 0xff)                    // 16 Bit-Wert
  {
    b= (h >> 12);
    hexnibbleout(b);
    b= (h >> 8) & 0x0f;
    hexnibbleout(b);
  }
  b= h;
  b= (h >> 4) & 0x0f;
  hexnibbleout(b);
  b= h & 0x0f;
  hexnibbleout(b);
}

void putstring(char *p)
{
  do
  {
    uart_putchar( *p );
  } while( *p++);
}


/* ------------------------------------------------------------
                             MY_PRINTF
     alternativer Ersatz fuer printf.

     Aufruf:

         my_printf("Ergebnis= %d",zahl);

     Platzhalterfunktionen:

        %s     : Ausgabe Textstring
        %d     : dezimale Ausgabe
        %x     : hexadezimale Ausgabe
                 ist Wert > 0xff erfolgt 4-stellige
                 Ausgabe
                 is Wert <= 0xff erfolgt 2-stellige
                 Ausgabe
        %k     : Integerausgabe als Pseudokommazahl
                 12345 wird als 123.45 ausgegeben
        %c     : Ausgabe als Asciizeichen

   ------------------------------------------------------------ */
void my_printf(const char *s,...)
{
  int       arg1;
  char      *arg2;
  uint8_t   token;
  char      ch;
  va_list   ap;

  va_start(ap,s);
  do
  {
    ch= *s;
    if(ch== 0) return;

    if(ch=='%')            // Platzhalterzeichen
    {
      s++;
      token= *s;
      switch(token)
      {
        case 'd':          // dezimale Ausgabe
        {
          arg1= va_arg(ap,int);
          putint(arg1,0);
          break;
        }
        case 'x':          // hexadezimale Ausgabe
        {
          arg1= va_arg(ap,int);
          puthex(arg1);
          break;
        }
        case 'k':
        {
          arg1= va_arg(ap,int);
          putint(arg1,printfkomma);     // Integerausgabe mit Komma: 12896 zeigt 12.896 an
          break;
        }
        case 'c':          // Zeichenausgabe
        {
          arg1= va_arg(ap,int);
          uart_putchar(arg1);
          break;
        }
        case '%':
        {
          uart_putchar(token);
          break;
        }
        case 's':
        {
          arg2= va_arg(ap,char *);
          putstring(arg2);
          break;
        }
      }
    }
    else
    {
      uart_putchar(ch);
    }
    s++;
  }while (ch != '\0');
}

